# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2018 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Core Data Stuff
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa

from rattail.core import get_uuid


def uuid_column(*args, **kwargs):
    """
    Returns a UUID column for use as a table's primary key.
    """

    kwargs.setdefault('primary_key', True)
    kwargs.setdefault('nullable', False)
    kwargs.setdefault('default', get_uuid)
    return sa.Column(sa.String(length=32), *args, **kwargs)


def filename_column(*args, **kwargs):
    """
    Returns a SQLAlchemy Column object suitable for representing a filename.
    """
    kwargs.setdefault('nullable', True)
    kwargs.setdefault('doc', "Base name of the data file.")
    return sa.Column(sa.String(length=255), *args, **kwargs)


def getset_factory(collection_class, proxy):
    """
    Get/set factory for SQLAlchemy association proxy attributes.
    """
    def getter(obj):
        if obj is None:
            return None
        return getattr(obj, proxy.value_attr)

    def setter(obj, val):
        setattr(obj, proxy.value_attr, val)

    return getter, setter



# -*- coding: utf-8; -*-
"""add label_batch.label_profile

Revision ID: 95a66a59d420
Revises: bd7acb7028da
Create Date: 2018-12-18 18:20:10.722491

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '95a66a59d420'
down_revision = 'bd7acb7028da'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # label_batch
    op.add_column('label_batch', sa.Column('label_profile_uuid', sa.String(length=32), nullable=True))
    op.create_foreign_key('label_batch_fk_label_profile', 'label_batch', 'label_profile', ['label_profile_uuid'], ['uuid'])


def downgrade():

    # label_batch
    op.drop_constraint('label_batch_fk_label_profile', 'label_batch', type_='foreignkey')
    op.drop_column('label_batch', 'label_profile_uuid')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Vendor catalog parser for KeHE Distributors
"""

from __future__ import unicode_literals, absolute_import

import os
import re
import datetime
import decimal
import logging

import xlrd
from sqlalchemy.orm import joinedload

from rattail.db import model
from rattail.db.cache import cache_model
from rattail.gpc import GPC
from rattail.vendors.catalogs import CatalogParser


log = logging.getLogger(__name__)


class KeheCatalogParser(CatalogParser):
    """
    Vendor catalog parser for KeHE Distributors.
    """
    key = 'rattail.contrib.kehe'
    display = "KeHE Distributors"
    vendor_key = 'kehe'

    def parse_effective_date(self, path):
        """
        This parser does not expect to find an effective date within the
        catalog file itself, but it makes a feeble attempt at gleaning it from
        the filename.  The example used to create this code was:

        .. code-block:: none

           DC22 CNP Catalog - February 2015.xls
        """
        basename = os.path.basename(path)
        match = re.search(r' (\w{3} \d{4})\.xls$', basename)
        if match:
            return datetime.datetime.strptime(match.group(1), '%b %Y').date()

    def parse_rows(self, path, progress=None):
        book = xlrd.open_workbook(path)
        sheet = book.sheet_by_index(0)
        products = cache_model(self.session, model.Product, key='upc',
                               query_options=[joinedload(model.Product.costs),
                                              joinedload(model.Product.cost)])

        for r in range(1, sheet.nrows): # Skip first header row.

            row = model.VendorCatalogBatchRow()
            upc = sheet.cell_value(r, 5) or None
            row.upc = GPC(int(upc)) if upc else None
            row.brand_name = sheet.cell_value(r, 1)
            row.description = sheet.cell_value(r, 2)
            row.size = sheet.cell_value(r, 3)
            row.vendor_code = '{0:07d}'.format(int(sheet.cell_value(r, 0)))

            # KeHE catalogs sort of have two columns to represent case size.
            # Sometimes the "Case Pack" column is what we want; other times
            # it's the "Un of Sls" column.  There appears to be no way to
            # really know which should be used, so we must query the database
            # to determine which case size is already in effect for the
            # product.  If that doesn't work, then we're back to guessing; we
            # use the "Case Pack" column in that case.
            case_pack = int(sheet.cell_value(r, 6))
            unit_of_sales = int(sheet.cell_value(r, 7))
            if row.upc:
                product = products.get(row.upc)
                if product:

                    # First we try to find an existing cost record for KeHE.  If
                    # one exists, we will honor its case size.
                    cost = product.cost_for_vendor(self.vendor)
                    if cost:
                        if cost.case_size == case_pack:
                            row.case_size = case_pack
                        elif cost.case_size == unit_of_sales:
                            row.case_size = unit_of_sales

                    # If there was no KeHE cost record, then we'll honor the
                    # "preferred" cost record case size, if it has one.
                    if row.case_size is None and product.cost:
                        if product.cost.case_size == case_pack:
                            row.case_size = case_pack
                        elif product.cost.case_size == unit_of_sales:
                            row.case_size = unit_of_sales

            # If all else fails, use the "Case Pack" column.
            if row.case_size is None:
                row.case_size = case_pack

            try:
                row.unit_cost = decimal.Decimal('{:0.4f}'.format(sheet.cell_value(r, 8)))
            except ValueError:
                log.warning("invalid unit cost for UPC {} at row {}: {}".format(row.upc, r + 1, sheet.cell_value(r, 8)))
            else:
                row.case_cost = row.unit_cost * row.case_size
            yield row



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Base classes for importers which have file(s) as source data
"""

from __future__ import unicode_literals, absolute_import

import os

from rattail import importing
from rattail.excel import ExcelReader


class FromFile(importing.Importer):
    """
    Base class for importers which obtain source (host) data from file.
    """

    def setup(self):
        if not hasattr(self, 'input_file_path'):
            self.input_file_path = self.get_input_file_path()
        self.open_input_file()

    def teardown(self):
        self.close_input_file()

    def get_input_file_path(self):
        return os.path.join(self.input_dir,
                            self.get_input_file_name())

    def get_input_file_name(self):
        raise NotImplementedError

    def open_input_file(self):
        raise NotImplementedError

    def close_input_file(self):
        raise NotImplementedError


class FromExcelFile(FromFile):
    """
    Base class for importers which get their data from Excel file.
    """

    def get_input_file_name(self):
        return '{}.xlsx'.format(self.model_name)

    def open_input_file(self):
        self.excel_reader = ExcelReader(self.input_file_path)

    def close_input_file(self):
        pass

    def get_host_objects(self):
        return self.excel_reader.read_rows()



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Logging Utilities
"""

from __future__ import unicode_literals, absolute_import

import os
import sys
import copy
import datetime
import socket
import getpass
import logging

import pytz

from rattail.time import localtime


class RattailAdapter(logging.LoggerAdapter):
    """
    Adds various items to a logger's context.

    Specifically, this adds the following keys to the ``extra`` dictionary
    available to the logger's formatter(s):

    * ``'hostname'`` - The fully-qualified domain name of the host machine.
    * ``'hostip'`` - The IPv4 address of the host machine.
    * ``'argv'`` - The value of ``sys.argv`` (a list).
    * ``'uid'`` - The effective UID of the running process.
    * ``'username'`` - The login name of the effective user.
    """

    def __init__(self, logger):
        hostname = socket.getfqdn()
        extra = {
            'hostname': hostname,
            'hostip':   socket.gethostbyname(hostname),
            'argv':     sys.argv,
            'uid':      "??",
            'username': getpass.getuser(),
            }
        if hasattr(os, 'getuid'):
            extra['uid'] = os.getuid()
        # LoggerAdapter is a new-style class only as of Python 2.7; must not
        # use super() in case we're running on Python 2.6.
        logging.LoggerAdapter.__init__(self, logger, extra)

    def process(self, msg, kwargs):
        extra = dict(self.extra)
        extra.update(kwargs.get('extra', {}))
        kwargs['extra'] = extra
        return msg, kwargs


def get_logger(name):
    """
    Fetches a logger by name, and wraps it in an adapter.

    This is intended to effectively replace the ``logging.getLogger()``
    function.  It does the same thing, except for wrapping the logger within a
    :class:`RattailAdapter`.
    """
    log = logging.getLogger(name)
    return RattailAdapter(log)


class TimeConverter(object):
    """
    Custom time converter for use with logging formatters.  Makes sure the
    final effective time zone for logged timestamps is the "default" one.
    """

    def __init__(self, config):
        self.config = copy.copy(config)
        self.config.usedb = False

    def __call__(self, value):
        value = datetime.datetime.utcfromtimestamp(value)
        value = pytz.utc.localize(value)
        value = localtime(self.config, value)
        return value.timetuple()


class LuigiSummaryFilter(logging.Filter):
    """
    Custom logging filter, to elevate the Luigi "execution summary" message to
    WARNING level, so that it shows up even when we ignore INFO logging.
    """

    def filter(self, record):
        if record.name == 'luigi-interface' and record.levelno == logging.INFO:
            if "===== Luigi Execution Summary =====" in record.msg:
                record.levelno = logging.WARNING
                record.levelname = 'WARNING'
        return True



# -*- coding: utf-8 -*-
"""add vendor schedule fields

Revision ID: 3f0d2a97f12a
Revises: 3c33f288ad62
Create Date: 2015-02-27 01:08:19.644637

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '3f0d2a97f12a'
down_revision = u'3c33f288ad62'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # vendor.lead_time_days
    op.add_column('vendor', sa.Column('lead_time_days', sa.Numeric(precision=5, scale=1), nullable=True))
    op.add_column('vendor_version', sa.Column('lead_time_days', sa.Numeric(precision=5, scale=1), autoincrement=False, nullable=True))

    # vendor.order_interval_days
    op.add_column('vendor', sa.Column('order_interval_days', sa.Numeric(precision=5, scale=1), nullable=True))
    op.add_column('vendor_version', sa.Column('order_interval_days', sa.Numeric(precision=5, scale=1), autoincrement=False, nullable=True))


def downgrade():

    # vendor.order_interval_days
    op.drop_column('vendor_version', 'order_interval_days')
    op.drop_column('vendor', 'order_interval_days')

    # vendor.lead_time_days
    op.drop_column('vendor_version', 'lead_time_days')
    op.drop_column('vendor', 'lead_time_days')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
People Handler

See also :doc:`rattail-manual:base/handlers/other/people`.
"""

from __future__ import unicode_literals, absolute_import

import warnings

import six
import sqlalchemy as sa

from rattail.app import GenericHandler
from rattail.time import make_utc


class PeopleHandler(GenericHandler):
    """
    Base class and default implementation for people handlers.
    """

    def normalize_full_name(self, first, last, **kwargs):
        """
        Normalize a "full" name based on the given first and last
        names.  Tries to be smart about collapsing whitespace etc.

        :param first: First name.
        :param last: Last name.
        :returns: First and last name combined.
        """
        from rattail.db.util import normalize_full_name
        return normalize_full_name(first, last)

    def make_person(self, **kwargs):
        """
        Make and return a new Person instance.
        """
        model = self.model
        person = model.Person()

        if 'first_name' in kwargs:
            person.first_name = kwargs.pop('first_name')
        if 'middle_name' in kwargs:
            person.middle_name = kwargs.pop('middle_name')
        if 'last_name' in kwargs:
            person.last_name = kwargs.pop('last_name')

        if 'display_name' in kwargs:
            person.display_name = kwargs.pop('display_name')
        else:
            person.display_name = self.normalize_full_name(
                person.first_name, person.last_name)

        for key, value in six.iteritems(kwargs):
            if hasattr(person, key):
                setattr(person, key, value)

        return person

    def update_names(self, person, **kwargs):
        """
        Update name(s) for the given person.

        :param person: Reference to a ``Person`` record.
        :param first: First name for the person.
        :param middle: Middle name for the person.
        :param last: Last name for the person.
        :param full: Full (display) name for the person.
        """
        if 'first' in kwargs:
            person.first_name = kwargs['first']

        if 'middle' in kwargs:
            person.middle_name = kwargs['middle']

        if 'last' in kwargs:
            person.last_name = kwargs['last']

        if 'full' in kwargs:
            if kwargs['full']:
                person.display_name = kwargs['full']
            else:
                person.display_name = self.normalize_full_name(
                    person.first_name, person.last_name)
        elif 'first' in kwargs and 'last' in kwargs:
            person.display_name = self.normalize_full_name(
                person.first_name, person.last_name)

    def add_phone(self, person, number, type='Home', preferred=False, **kwargs):
        """
        Add a phone record for the person.

        :param person: Reference to a ``Person`` record.
        :param number: Actual phone number to add.
        :param type: Type of phone number to add.
        :param preferred: Boolean indicating that this should be the
           new "preferred" number for the person.
        """
        reason = self.app.phone_number_is_invalid(number)
        if reason:
            raise ValueError("Phone number is not valid: {}".format(reason))

        phone = person.add_phone(number=self.app.format_phone_number(number),
                                 type=type)
        if preferred:
            person.set_primary_phone(phone)

        return phone

    def update_phone(self, person, phone, **kwargs):
        """
        Update a phone record for the person.

        :param person: Reference to a ``Person`` record.
        :param phone: Reference to the ``PersonPhoneNumber`` record to update.
        :param number: Actual phone number.
        :param type: Type of phone number.
        :param preferred: Boolean indicating that this should be the
           new "preferred" number for the person.
        """
        if phone not in person.phones:
            raise ValueError("Phone does not belong to this person.")

        if 'number' in kwargs:
            reason = self.app.phone_number_is_invalid(kwargs['number'])
            if reason:
                raise ValueError("Phone number is not valid: {}".format(reason))
            phone.number = self.app.format_phone_number(kwargs['number'])

        if 'type' in kwargs:
            phone.type = kwargs['type']

        if 'preferred' in kwargs:
            if kwargs['preferred']:
                person.set_primary_phone(phone)
            else: # should *not* prefer this one
                if phone.preferred and len(person.phones) > 1:
                    # make 2nd phone the new 1st
                    person.set_primary_phone(person.phones[1])

        return phone

    def add_email(self, person, address, type='Home',
                  invalid=False, preferred=False, **kwargs):
        """
        Add a email record for the person.

        :param person: Reference to a ``Person`` record.
        :param address: Actual email address to add.
        :param type: Type of email address to add.
        :param invalid: Boolean indicating the address is known to
           *not* be valid.
        :param preferred: Boolean indicating that this should (or not)
           be the "preferred" email address for the person.
        """
        email = person.add_email(address=address,
                                 type=type,
                                 invalid=invalid)
        if preferred:
            person.set_primary_email(email)

        return email

    def update_email(self, person, email, **kwargs):
        """
        Update a email record for the person.

        :param person: Reference to a ``Person`` record.
        :param email: Reference to the ``PersonEmailAddress`` record to update.
        :param address: Actual email address.
        :param type: Type of email address.
        :param invalid: Boolean indicating the address is known to
           *not* be valid.
        :param preferred: Boolean indicating that this should (or not)
           be the "preferred" address for the person.
        """
        if email not in person.emails:
            raise ValueError("Email does not belong to this person.")

        if 'address' in kwargs:
            email.address = kwargs['address']

        if 'type' in kwargs:
            email.type = kwargs['type']

        if 'invalid' in kwargs:
            email.invalid = kwargs['invalid']

        if 'preferred' in kwargs:
            if kwargs['preferred']:
                person.set_primary_email(email)
            else: # should *not* prefer this one
                if email.preferred and len(person.emails) > 1:
                    # make 2nd email the new 1st
                    person.set_primary_email(person.emails[1])

        return email

    def ensure_address(self, person, **kwargs):
        """
        Returns the default address record associated with the given
        person, creating it first if necessary.
        """
        address = person.first_address()
        if not address:
            address = person.add_address(**kwargs)
            # TODO: this might be a good idea..maybe if a kwarg flag is set?
            # person.set_primary_address(address)
        return address

    def address_is_invalid(self, person, address, **kwargs):
        """
        Returns a boolean indicating if the given person's address is
        invalid.

        :param person: Reference to a person.
        :param address: Reference to a person's address.
        """
        return address.invalid

    def mark_address_invalid(self, person, address, invalid, **kwargs):
        """
        Mark the person's address as invalid.

        :param person: Reference to a person.
        :param address: Reference to a person's address.
        :param invalid: Boolean indicating "invalid" status for
           person's address.
        """
        address.invalid = invalid

    def request_merge(self, user, removing_uuid, keeping_uuid, **kwargs):
        """
        Submit an officical merge request for two Person records.

        The caller must obviously specify which is to be kept and
        which removed, but really this is arbitrary, as the user
        performing the merge is free to swap them around.
        """
        model = self.model
        session = self.get_session(user)
        merge = model.MergePeopleRequest()
        merge.removing_uuid = removing_uuid
        merge.keeping_uuid = keeping_uuid
        merge.requested_by = user
        merge.requested = make_utc()
        session.add(merge)
        session.flush()
        self.notify_of_merge_request(merge)
        return merge

    def notify_of_merge_request(self, merge):
        """
        Send an email alert regarding a new merge request.
        """
        session = self.get_session(merge)
        model = self.model

        removing = session.query(model.Person).get(merge.removing_uuid)
        keeping = session.query(model.Person).get(merge.keeping_uuid)

        context = {
            'user_display': merge.requested_by.display_name,
            'removing_display': six.text_type(removing) if removing else "(not found)",
            'keeping_display': six.text_type(keeping) if keeping else "(not found)",
        }

        url = self.config.base_url()
        if url:
            context['merge_request_url'] = '{}/people/merge-requests/{}'.format(url, merge.uuid)
            if removing:
                context['removing_url'] = '{}/people/{}/profile'.format(url, removing.uuid)
            if keeping:
                context['keeping_url'] = '{}/people/{}/profile'.format(url, keeping.uuid)

        self.app.send_email('person_merge_request', context)

    def get_merge_preview_fields(self, **kwargs):
        """
        Returns a sequence of fields which will be used during a merge
        preview.
        """
        def F(name, **kwargs):
            field = {'name': name}
            field.update(kwargs)
            return field

        return [
            F('uuid'),
            F('first_name'),
            F('last_name'),
            F('display_name'),
            F('usernames', additive=True),
            F('employee_uuid', coalesce=True),
            F('member_uuids', additive=True),
        ]

    def get_merge_preview_data(self, person, **kwargs):
        """
        Must return a data dictionary for the given person, which can
        be presented to the user during a merge preview.
        """
        return {
            'uuid': person.uuid,
            'first_name': person.first_name,
            'last_name': person.last_name,
            'display_name': person.display_name,
            'usernames': [u.username for u in person.users],
            'employee_uuid': person.employee.uuid if person.employee else None,
            'member_uuids': [m.uuid for m in person.members],
        }

    def get_merge_resulting_data(self, removing, keeping, **kwargs):
        """
        Must return a dictionary to represent what the *final* data
        would look like, should the proposed merge occur.  Note that
        we're still in preview mode here, this doesn't actually cause
        any particular data to become final.

        :param removing: Data dictionary for the Person to be removed,
           as obtained via :meth:`get_merge_preview_data()`.
        :param keeping: Data dictionary for the Person to be preserved,
           as obtained via :meth:`get_merge_preview_data()`.
        """
        fields = self.get_merge_preview_fields()
        coalesce_fields = [f for f in fields
                           if f.get('coalesce')]
        additive_fields = [f for f in fields
                           if f.get('additive')]

        # start with clone of the `keeping` dict
        result = dict(keeping)

        # coalesce any field values which need it
        for field in coalesce_fields:
            if removing[field] is not None and keeping[field] is None:
                result[field] = removing[field]
            elif removing[field] and not keeping[field]:
                result[field] = removing[field]

        # sum any field values which need it
        for field in additive_fields:
            if isinstance(keeping[field], (list, tuple)):
                result[field] = sorted(set(removing[field] + keeping[field]))
            else:
                result[field] = removing[field] + keeping[field]

        return result

    def why_not_merge(self, removing, keeping, **kwargs):
        """
        Evaluate the given merge candidates and if there is a reason *not*
        to merge them, return that reason.

        :param removing: Person record which will be removed, should the
           merge happen.
        :param keeping: Person record which will be kept, should the
           merge happen.
        :returns: String indicating reason not to merge, or ``None``.
        """
        if removing.employee and keeping.employee:
            if removing.employee is not keeping.employee:
                return "Cannot merge 2 people who are distinct employees"

    def perform_merge(self, removing, keeping, **kwargs):
        """
        Perform an actual merge of the 2 given people.

        :param removing: Person record which should be removed.
        :param keeping: Person record which should be kept.
        """
        # move Member records to final Person
        for member in list(removing.members):
            removing.members.remove(member)
            keeping.members.append(member)

        # move Employee record to final Person
        if removing.employee:
            if keeping.employee and keeping.employee is not removing.employee:
                raise RuntimeError("Cannot merge 2 people who are distinct employees")
            if not keeping.employee:
                employee = removing.employee
                employee.person = keeping

        # move User records to final Person
        for user in list(removing.users):
            removing.users.remove(user)
            keeping.users.append(user)

        # delete unwanted Person
        session = self.get_session(keeping)
        session.delete(removing)
        session.flush()

        self.satisfy_merge_requests(removing, keeping, user=kwargs.get('user'))

    def satisfy_merge_requests(self, removing, keeping, user):
        """
        If there was a merge request(s) for this pair, mark it complete.
        """
        session = self.get_session(keeping)
        model = self.model
        merge_requests = session.query(model.MergePeopleRequest)\
                                .filter(sa.or_(
                                    sa.and_(
                                        model.MergePeopleRequest.removing_uuid == removing.uuid,
                                        model.MergePeopleRequest.keeping_uuid == keeping.uuid),
                                    sa.and_(
                                        model.MergePeopleRequest.removing_uuid == keeping.uuid,
                                        model.MergePeopleRequest.keeping_uuid == removing.uuid)))\
                                .all()
        for merge_request in merge_requests:
            # set the record straight re: removing vs. keeping
            merge_request.removing_uuid = removing.uuid
            merge_request.keeping_uuid = keeping.uuid
            merge_request.merged = make_utc()
            merge_request.merged_by = user

def get_people_handler(config, **kwargs):
    """
    Create and return the configured :class:`PeopleHandler` instance.

    .. warning::
       This function is deprecated; please use
       :meth:`~rattail.app.AppHandler.get_people_handler` instead.
    """
    warnings.warn("get_people_handler() function is deprecated, "
                  "please use app.get_people_handler() method instead",
                  DeprecationWarning)
    app = config.get_app()
    return app.get_people_handler(**kwargs)



# -*- coding: utf-8 -*-
"""add purchase_batch.invoice_file

Revision ID: 92a49edc45c3
Revises: 6551a4d9ff25
Create Date: 2018-05-16 13:46:31.940108

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = '92a49edc45c3'
down_revision = u'6551a4d9ff25'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # purchase_batch
    op.add_column('purchase_batch', sa.Column('invoice_file', sa.String(length=255), nullable=True))
    op.add_column('purchase_batch', sa.Column('invoice_parser_key', sa.String(length=100), nullable=True))


def downgrade():

    # purchase_batch
    op.drop_column('purchase_batch', 'invoice_parser_key')
    op.drop_column('purchase_batch', 'invoice_file')



# -*- coding: utf-8; -*-
"""add importer batch

Revision ID: b82daacc86b7
Revises: a809caf23cf0
Create Date: 2017-12-29 00:16:43.897114

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'b82daacc86b7'
down_revision = u'0c91cf7d557b'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # batch_importer
    op.create_table('batch_importer',
                    sa.Column('uuid', sa.String(length=32), nullable=False),
                    sa.Column('id', sa.Integer(), nullable=False),
                    sa.Column('description', sa.String(length=255), nullable=True),
                    sa.Column('created', sa.DateTime(), nullable=False),
                    sa.Column('created_by_uuid', sa.String(length=32), nullable=False),
                    sa.Column('cognized', sa.DateTime(), nullable=True),
                    sa.Column('cognized_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('rowcount', sa.Integer(), nullable=True),
                    sa.Column('complete', sa.Boolean(), nullable=True),
                    sa.Column('executed', sa.DateTime(), nullable=True),
                    sa.Column('executed_by_uuid', sa.String(length=32), nullable=True),
                    sa.Column('purge', sa.Date(), nullable=True),
                    sa.Column('notes', sa.Text(), nullable=True),
                    sa.Column('status_code', sa.Integer(), nullable=True),
                    sa.Column('status_text', sa.String(length=255), nullable=True),
                    sa.Column('row_table', sa.String(length=255), nullable=False),
                    sa.Column('batch_handler_spec', sa.String(length=255), nullable=True),
                    sa.Column('import_handler_spec', sa.String(length=255), nullable=False),
                    sa.Column('host_title', sa.String(length=255), nullable=False),
                    sa.Column('local_title', sa.String(length=255), nullable=False),
                    sa.Column('importer_key', sa.String(length=100), nullable=False),
                    sa.ForeignKeyConstraint(['cognized_by_uuid'], [u'user.uuid'], name=u'batch_importer_fk_cognized_by'),
                    sa.ForeignKeyConstraint(['created_by_uuid'], [u'user.uuid'], name=u'batch_importer_fk_created_by'),
                    sa.ForeignKeyConstraint(['executed_by_uuid'], [u'user.uuid'], name=u'batch_importer_fk_executed_by'),
                    sa.PrimaryKeyConstraint('uuid')
    )


def downgrade():

    # batch_importer
    op.drop_table('batch_importer')



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2021 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Autocomplete Handler for Products
"""

from __future__ import unicode_literals, absolute_import

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.autocomplete import Autocompleter
from rattail.db import model


class ProductAutocompleter(Autocompleter):
    """
    Autocompleter for Products.

    Actually this will search both the
    :attr:`~rattail.db.model.products.Brand.name` and
    :attr:`~rattail.db.model.products.Product.description` fields.

    Note that this will *not* include products marked as "deleted" -
    see :class:`ProductAllAutocompleter` if you need those too for
    some reason.
    """
    autocompleter_key = 'products'
    model_class = model.Product
    autocomplete_fieldname = 'description'

    def make_base_query(self, session):
        model = self.model
        return session.query(model.Product)\
                      .outerjoin(model.Brand)\
                      .options(orm.joinedload(model.Product.brand))

    def restrict_autocomplete_query(self, session, query, **kwargs):
        model = self.model

        # do not show "deleted" items by default
        query = query.filter(model.Product.deleted == False)

        return query

    def filter_autocomplete_query(self, session, query, term):
        model = self.model

        # filter by user-provided term
        criteria = []
        for word in term.split():
            criteria.append(sa.or_(
                model.Brand.name.ilike('%{}%'.format(word)),
                model.Product.description.ilike('%{}%'.format(word))))
        query = query.filter(sa.and_(*criteria))

        return query

    def sort_autocomplete_query(self, session, query):
        model = self.model
        return query.order_by(model.Brand.name,
                              model.Product.description)

    def autocomplete_display(self, product):
        return product.full_description


class ProductAllAutocompleter(ProductAutocompleter):
    """
    Autocompleter for Products, which shows *all* results, including
    "deleted" items etc.
    """

    def restrict_autocomplete_query(self, session, query, **kwargs):
        return query


class ProductNewOrderAutocompleter(ProductAutocompleter):
    """
    Special "new order" autocompleter for products.

    We set it apart with a different key (``'products.neworder'``) so
    that you can override it independently of other product
    autocompleters.

    But the default logic for this one is a bit special too, in that
    it will try to search for *both* UPC (or whatever your "key" is)
    *and* product brand, description etc.  The UPC (key) must match
    "exactly" whereas the description etc. uses wildcard search.

    .. attribute:: lookup_only_keys

       You can set this to an explicit list of product "keys" to
       restrict which types of lookups are performed for that.

     """
    autocompleter_key = 'products.neworder'
    model_class = model.Product
    autocomplete_fieldname = 'description'
    lookup_only_keys = []

    def get_autocomplete_data(self, session, term, **kwargs):
        """
        Collect data for all matching results.  This will run two
        queries, one for product "key" (e.g. UPC) and another to
        search brand, description etc.

        Note that this still honors :attr:`max_results` and will run
        the product "key" query first, without limit since it will
        have very few results.  Then the wildcard search query runs
        and will be limited.
        """
        # first run the product "key" query, and save matches for later
        self.key_matches = self.find_product_key_matches(session, term,
                                                         include_keys=True)

        # then run the "normal" upstream query to get data from the
        # wildcard search.  note that this will be results-limited
        data = super(ProductNewOrderAutocompleter, self).get_autocomplete_data(
            session, term, **kwargs)

        return data

    def get_autocomplete_results(self, data):
        """
        Format the data into a final results set for return to the
        caller.
        """
        # first get normal "search" results
        data = super(ProductNewOrderAutocompleter, self).get_autocomplete_results(data)

        # then inject key matches at the beginning
        for i, match in enumerate(self.key_matches):
            key, product = match
            data.insert(i, {'label': self.autocomplete_display(product),
                            'value': self.autocomplete_value(product)})

        return data

    def find_product_key_matches(self, session, term, **kwargs):
        """
        Find the products where the "key" matches the given term.

        Usually this just means finding a UPC match.
        """
        handler = self.app.get_products_handler()
        if 'only' not in kwargs and self.lookup_only_keys:
            kwargs['only'] = self.lookup_only_keys
        products = handler.find_products_by_key(session, term, **kwargs)
        return products



# -*- coding: utf-8; -*-
"""add custorder_batch_row.product_upc

Revision ID: 2afee42cc24d
Revises: a7b286e87f89
Create Date: 2021-01-25 21:32:00.008209

"""

from __future__ import unicode_literals

# revision identifiers, used by Alembic.
revision = '2afee42cc24d'
down_revision = 'a7b286e87f89'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types



def upgrade():

    # custorder
    op.add_column('custorder', sa.Column('total_price', sa.Numeric(precision=10, scale=3), nullable=True))

    # custorder_item
    op.alter_column('custorder_item', 'product_unit_of_measure',
               existing_type=sa.VARCHAR(length=4),
               nullable=True)
    op.add_column('custorder_item', sa.Column('product_upc', rattail.db.types.GPCType(), nullable=True))
    op.add_column('custorder_item', sa.Column('product_weighed', sa.Boolean(), nullable=True))
    op.add_column('custorder_item', sa.Column('order_quantity', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('custorder_item', sa.Column('order_uom', sa.String(length=4), nullable=True))

    # custorder_batch
    op.add_column('custorder_batch', sa.Column('total_price', sa.Numeric(precision=10, scale=3), nullable=True))

    # custorder_batch_row
    op.alter_column('custorder_batch_row', 'product_unit_of_measure',
               existing_type=sa.VARCHAR(length=4),
               nullable=True)
    op.add_column('custorder_batch_row', sa.Column('product_upc', rattail.db.types.GPCType(), nullable=True))
    op.add_column('custorder_batch_row', sa.Column('product_weighed', sa.Boolean(), nullable=True))
    op.add_column('custorder_batch_row', sa.Column('order_quantity', sa.Numeric(precision=10, scale=4), nullable=True))
    op.add_column('custorder_batch_row', sa.Column('order_uom', sa.String(length=4), nullable=True))


def downgrade():

    # custorder_batch_row
    op.drop_column('custorder_batch_row', 'order_uom')
    op.drop_column('custorder_batch_row', 'order_quantity')
    op.drop_column('custorder_batch_row', 'product_weighed')
    op.drop_column('custorder_batch_row', 'product_upc')
    # TODO: this is really a one-way change, cannot be easily undone
    # op.alter_column('custorder_batch_row', 'product_unit_of_measure',
    #            existing_type=sa.VARCHAR(length=4),
    #            nullable=False)

    # custorder_batch
    op.drop_column('custorder_batch', 'total_price')

    # custorder_item
    op.drop_column('custorder_item', 'order_uom')
    op.drop_column('custorder_item', 'order_quantity')
    op.drop_column('custorder_item', 'product_weighed')
    op.drop_column('custorder_item', 'product_upc')
    # TODO: this is really a one-way change, cannot be easily undone
    # op.alter_column('custorder_item', 'product_unit_of_measure',
    #            existing_type=sa.VARCHAR(length=4),
    #            nullable=False)

    # custorder
    op.drop_column('custorder', 'total_price')



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Shopfoo model importers
"""

import shutil


class ProductImporterMixin(object):
    """
    Exporter target for Shopfoo product data
    """
    model_name = 'Product'
    allow_delete = False

    @property
    def export_model_class(self):
        raise NotImplementedError("You must define {}.export_model_class".format(
            self.__class__.__name__))

    def init_export(self, session):
        self.current_export = self.export_model_class(created_by=self.runas_user)
        session.add(self.current_export)
        session.flush()
        session.refresh(self.current_export)
        self.current_export.set_filename(self.config)
        self.record_count = 0

    def finalize_export(self):
        export = self.current_export

        # preserve final item count in the export record itself
        export.record_count = self.record_count

        if not self.dry_run:

            # copy output file to export's data path
            path = self.config.export_filepath(export.export_key,
                                               export.uuid,
                                               export.filename,
                                               makedirs=True)
            shutil.copyfile(self.output_file_path, path)



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import shutil
import tempfile

from mock import patch, Mock

from rattail.importing import rattail_bulk as bulk
from rattail.tests.importing import ImporterTester
from rattail.tests.importing.test_rattail import DualRattailTestCase


class BulkImportTester(DualRattailTestCase, ImporterTester):

    handler_class = bulk.BulkFromRattailToRattail

    def setUp(self):
        self.setup_rattail()
        self.tempdir = tempfile.mkdtemp()
        self.config.set('rattail', 'workdir', self.tempdir)
        self.handler = self.make_handler()

        # TODO: no-op for coverage, how lame is that
        self.handler.get_default_keys()

    def tearDown(self):
        self.teardown_rattail()
        shutil.rmtree(self.tempdir)

    @property
    def model_name(self):
        return self.make_importer().model_name

    def get_fields(self):
        return self.make_importer().fields

    def make_handler(self, **kwargs):
        if 'config' not in kwargs and hasattr(self, 'config'):
            kwargs['config'] = self.config
        return self.handler_class(**kwargs)

    def import_data(self, host_data=None, **kwargs):
        if host_data is None:
            fields = self.get_fields()
            host_data = list(self.copy_data().values())
            for data in host_data:
                for field in fields:
                    data.setdefault(field, None)
        with patch.object(self.importer_class, 'normalize_host_data', Mock(return_value=host_data)):
            with patch.object(self.handler, 'make_host_session', Mock(return_value=self.host_session)):
                return self.handler.import_data(self.model_name, **kwargs)


class TestPersonImport(BulkImportTester):

    importer_class = bulk.PersonImporter

    sample_data = {
        'fred': {
            'uuid': 'fred',
            'first_name': 'Fred',
            'last_name': 'Flintstone',
        },
        'maurice': {
            'uuid': 'maurice',
            'first_name': 'Maurice',
            'last_name': 'Jones',
        },
        'zebra': {
            'uuid': 'zebra',
            'first_name': 'Zebra',
            'last_name': 'Jones',
        },
    }

    def test_create(self):
        if self.postgresql():
            result = self.import_data()
            self.assertEqual(result, {'Person': 3})

    def test_max_create(self):
        if self.postgresql():
            result = self.import_data(max_create=1)
            self.assertEqual(result, {'Person': 1})


class TestProductImport(BulkImportTester):

    importer_class = bulk.ProductImporter

    def test_simple_fields(self):
        importer = self.make_importer()
        self.assertNotIn('regular_price_uuid', importer.simple_fields)
        self.assertNotIn('current_price_uuid', importer.simple_fields)



# -*- coding: utf-8; -*-
"""grow product.description

Revision ID: aa32b6e51129
Revises: 8e965902bf93
Create Date: 2017-03-23 18:18:02.286623

"""

from __future__ import unicode_literals, absolute_import

# revision identifiers, used by Alembic.
revision = 'aa32b6e51129'
down_revision = u'8e965902bf93'
branch_labels = None
depends_on = None

from alembic import op
import sqlalchemy as sa
import rattail.db.types
from sqlalchemy.dialects import postgresql


def upgrade():

    # product
    op.alter_column('product', 'description', type_=sa.String(length=255))
    op.alter_column('product', 'description2', type_=sa.String(length=255))


def downgrade():

    # product
    op.alter_column('product', 'description', type_=sa.String(length=60))
    op.alter_column('product', 'description2', type_=sa.String(length=60))



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2020 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
IFPS -> Rattail data import
"""

from __future__ import unicode_literals, absolute_import

import datetime
import logging

import sqlalchemy as sa

from rattail import importing
from rattail.importing.handlers import FromFileHandler
from rattail.importing.files import FromExcelFile
from rattail.util import OrderedDict
from rattail.time import localtime, make_utc
from rattail.db.util import maxlen


log = logging.getLogger(__name__)


class FromIFPSToRattail(FromFileHandler, importing.ToRattailHandler):
    """
    Handler for IFPS -> Rattail data import.
    """
    host_title = "IFPS"
    local_title = "Rattail"

    def get_importers(self):
        importers = OrderedDict()
        importers['IFPS_PLU'] = IFPS_PLUImporter
        return importers


class FromIFPS(FromExcelFile):
    """
    Base class for IFPS -> Rattail data importers.
    """


class IFPS_PLUImporter(FromIFPS, importing.model.IFPS_PLUImporter):
    """
    Imports PLU data from IFPS
    """
    key = 'plu'
    supported_fields = [
        'plu',
        'category',
        'commodity',
        'variety',
        'size',
        'measurements_north_america',
        'measurements_rest_of_world',
        'restrictions_notes',
        'botanical_name',
        'aka',
        'revision_date',
        'date_added',
        'gpc',
        'image',
        'image_source',
    ]

    def normalize_host_object(self, xlrow):
        plu = int(xlrow['PLU'])
        data = {
            'plu': plu,
            'category': xlrow['CATEGORY'].strip() or None,
            'commodity': xlrow['COMMODITY'].strip() or None,
            'variety': xlrow['VARIETY'].strip() or None,
            'size': xlrow['SIZE'].strip() or None,
            'measurements_north_america': xlrow['MEASUREMENTS: NORTH AMERICA'].strip() or None,
            'measurements_rest_of_world': xlrow['MEASUREMENTS: REST OF WORLD'].strip() or None,
            'restrictions_notes': xlrow['RESTRICTIONS / NOTES'].strip() or None,
            'botanical_name': xlrow['BOTANICAL NAME'].strip() or None,
            'aka': xlrow['AKA'].strip() or None,
            'gpc': xlrow['GPC'].strip() or None,
            'image': xlrow['IMAGE'].strip() or None,
            'image_source': xlrow['IMAGE_SOURCE'].strip() or None,
        }

        # TODO: these two date fields are being interpreted as if they were in
        # the *local* time zone, but then we of course convert them to UTC when
        # importing to Rattail DB.  however, the IFPS website and data do not
        # actually indicate which time zone should apply to their values.

        if 'revision_date' in self.fields:
            dt = datetime.datetime.strptime(xlrow['REVISION DATE'], '%Y-%m-%d %H:%M:%S')
            dt = localtime(self.config, dt)
            data['revision_date'] = make_utc(dt)

        if 'date_added' in self.fields:
            dt = datetime.datetime.strptime(xlrow['DATE ADDED'], '%Y-%m-%d %H:%M:%S')
            dt = localtime(self.config, dt)
            data['date_added'] = make_utc(dt)

        # i think our schema is good enough, but just in case this it's not,
        # this should help track down any lingering issues.
        for field in self.fields:
            column = getattr(self.model_class, field)
            if isinstance(column.type, sa.String):
                length = maxlen(column)
                if length and data[field] and len(data[field]) > length:
                    log.warning("field %s is too long (%s chars) for %s: %s",
                                field, len(data[field]), plu, data[field])
                    data[field] = data[field][:length]

        return data



# -*- coding: utf-8 -*-

from __future__ import unicode_literals, absolute_import

import datetime
import shutil
import tempfile
import unittest

import sqlalchemy as sa
from sqlalchemy import orm

from rattail.db import Session, model
from rattail.importing import postgresql as pgimport
from rattail.config import RattailConfig
from rattail.exceptions import ConfigurationError
from rattail.tests import RattailTestCase, NullProgress
from rattail.tests.importing import ImporterTester
from rattail.tests.importing.test_rattail import DualRattailTestCase
from rattail.time import localtime


class Widget(object):
    pass


class TestBulkToPostgreSQL(unittest.TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config = RattailConfig()
        self.config.set('rattail', 'workdir', self.tempdir)
        self.config.set('rattail', 'timezone.default', 'America/Chicago')

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def make_importer(self, **kwargs):
        kwargs.setdefault('config', self.config)
        kwargs.setdefault('fields', ['id']) # hack
        return pgimport.BulkToPostgreSQL(**kwargs)

    def test_data_path_property(self):
        self.config.set('rattail', 'workdir', '/tmp')
        importer = pgimport.BulkToPostgreSQL(config=self.config, fields=['id'])

        # path leverages model name, so default is None
        self.assertEqual(importer.data_path, '/tmp/import_bulk_postgresql_None.csv')

        # but it will reflect model name
        importer.model_name = 'Foo'
        self.assertEqual(importer.data_path, '/tmp/import_bulk_postgresql_Foo.csv')

    def test_setup(self):
        importer = self.make_importer()
        self.assertFalse(hasattr(importer, 'data_buffer'))
        importer.setup()
        self.assertIsNotNone(importer.data_buffer)
        importer.data_buffer.close()

    def test_teardown(self):
        importer = self.make_importer()
        importer.data_buffer = open(importer.data_path, 'wb')
        importer.teardown()
        self.assertIsNone(importer.data_buffer)

    def test_prep_value_for_postgres(self):
        importer = self.make_importer()

        # constants
        self.assertEqual(importer.prep_value_for_postgres(None), '\\N')
        self.assertEqual(importer.prep_value_for_postgres(True), 't')
        self.assertEqual(importer.prep_value_for_postgres(False), 'f')

        # datetime (local zone is Chicago/CDT; UTC-5)
        value = localtime(self.config, datetime.datetime(2016, 5, 13, 12))
        self.assertEqual(importer.prep_value_for_postgres(value), '2016-05-13 17:00:00')

        # strings...

        # backslash is escaped by doubling
        self.assertEqual(importer.prep_value_for_postgres('\\'), '\\\\')

        # newlines are collapsed (\r\n -> \n) and escaped
        self.assertEqual(importer.prep_value_for_postgres('one\rtwo\nthree\r\nfour\r\nfive\nsix\rseven'), 'one\\rtwo\\nthree\\r\\nfour\\r\\nfive\\nsix\\rseven')

    def test_prep_data_for_postgres(self):
        importer = self.make_importer()
        time = localtime(self.config, datetime.datetime(2016, 5, 13, 12))
        data = {
            'none': None,
            'true': True,
            'false': False,
            'datetime': time,
            'backslash': '\\',
            'newlines': 'one\rtwo\nthree\r\nfour\r\nfive\nsix\rseven',
        }
        data = importer.prep_data_for_postgres(data)
        self.assertEqual(data['none'], '\\N')
        self.assertEqual(data['true'], 't')
        self.assertEqual(data['false'], 'f')
        self.assertEqual(data['datetime'], '2016-05-13 17:00:00')
        self.assertEqual(data['backslash'], '\\\\')
        self.assertEqual(data['newlines'], 'one\\rtwo\\nthree\\r\\nfour\\r\\nfive\\nsix\\rseven')


######################################################################
# fake importer class, tested mostly for basic coverage
######################################################################

class MockBulkImporter(pgimport.BulkToPostgreSQL):
    model_class = model.Department
    key = 'uuid'

    def normalize_local_object(self, obj):
        return obj

    def update_object(self, obj, host_data, local_data=None):
        return host_data


class TestMockBulkImporter(DualRattailTestCase, ImporterTester):
    importer_class = MockBulkImporter

    sample_data = {
        1: {'number': 1, 'name': "Grocery", 'uuid': 'decd909a194011e688093ca9f40bc550'},
        2: {'number': 2, 'name': "Bulk", 'uuid': 'e633d54c194011e687e33ca9f40bc550'},
        3: {'number': 3, 'name': "HBA", 'uuid': 'e2bad79e194011e6a4783ca9f40bc550'},
    }

    def setUp(self):
        self.setup_rattail()
        self.tempdir = tempfile.mkdtemp()
        self.config.set('rattail', 'workdir', self.tempdir)
        self.importer = self.make_importer()

    def tearDown(self):
        self.teardown_rattail()
        shutil.rmtree(self.tempdir)

    def make_importer(self, **kwargs):
        kwargs.setdefault('config', self.config)
        return super(TestMockBulkImporter, self).make_importer(**kwargs)

    def import_data(self, **kwargs):
        self.importer.session = self.session
        self.importer.host_session = self.host_session
        self.result = self.importer.import_data(**kwargs)

    def assert_import_created(self, *keys):
        pass

    def assert_import_updated(self, *keys):
        pass

    def assert_import_deleted(self, *keys):
        pass

    def test_create(self):
        if self.postgresql():
            with self.host_data(self.sample_data):
                self.import_data()
            self.assert_import_created(3)

    def test_create_empty(self):
        if self.postgresql():
            with self.host_data({}):
                self.import_data()
            self.assert_import_created(0)

    def test_max_create(self):
        if self.postgresql():
            with self.host_data(self.sample_data):
                with self.local_data({}):
                    self.import_data(max_create=1)
            self.assert_import_created(1)

    def test_max_total_create(self):
        if self.postgresql():
            with self.host_data(self.sample_data):
                with self.local_data({}):
                    self.import_data(max_total=1)
            self.assert_import_created(1)

    # # TODO: a bit hacky, leveraging the fact that 'user' is a reserved word
    # def test_table_name_is_reserved_word(self):
    #     if self.postgresql():
    #         from rattail.importing.rattail_bulk import UserImporter
    #         data = {
    #             '521a788e195911e688c13ca9f40bc550': {
    #                 'uuid': '521a788e195911e688c13ca9f40bc550',
    #                 'username': 'fred',
    #                 'active': True,
    #             },
    #         }
    #         self.importer = UserImporter(config=self.config)
    #         # with self.host_data(data):
    #         self.import_data(host_data=data)
    #         # self.assert_import_created(3)



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import os
import csv
import datetime
import argparse
import shutil
import tempfile
from unittest import TestCase
from six import StringIO

import six
# from sqlalchemy import func
from mock import patch, Mock

from rattail.commands import core
from rattail.db import Session, model
from rattail.db.auth import authenticate_user
from rattail.tests import DataTestCase


class TestArgumentParser(TestCase):

    def test_parse_args_preserves_extra_argv(self):
        parser = core.ArgumentParser()
        parser.add_argument('--some-optional-arg')
        parser.add_argument('some_required_arg')
        args = parser.parse_args([
                '--some-optional-arg', 'optional-value', 'required-value',
                'some', 'extra', 'args'])
        self.assertEqual(args.some_required_arg, 'required-value')
        self.assertEqual(args.some_optional_arg, 'optional-value')
        self.assertEqual(args.argv, ['some', 'extra', 'args'])


class TestDateArgument(TestCase):

    def test_valid_date_string_returns_date_object(self):
        date = core.date_argument('2014-01-01')
        self.assertEqual(date, datetime.date(2014, 1, 1))

    def test_invalid_date_string_raises_error(self):
        self.assertRaises(argparse.ArgumentTypeError, core.date_argument, 'invalid-date')


class TestCommand(TestCase):

    def test_initial_subcommands_are_sane(self):
        command = core.Command()
        self.assertTrue('filemon' in command.subcommands)

    def test_unicode(self):
        command = core.Command()
        command.name = 'some-app'
        self.assertEqual(six.text_type(command), 'some-app')
        
    def test_iter_subcommands_includes_expected_item(self):
        command = core.Command()
        found = False
        for subcommand in command.iter_subcommands():
            if subcommand.name == 'filemon':
                found = True
                break
        self.assertTrue(found)

    def test_print_help(self):
        command = core.Command()
        stdout = StringIO()
        command.stdout = stdout
        command.print_help()
        output = stdout.getvalue()
        stdout.close()
        self.assertTrue('Usage:' in output)
        self.assertTrue('Options:' in output)

    def test_run_with_no_args_prints_help(self):
        command = core.Command()
        with patch.object(command, 'print_help') as print_help:
            command.run()
            print_help.assert_called_once_with()

    def test_run_with_single_help_arg_prints_help(self):
        command = core.Command()
        with patch.object(command, 'print_help') as print_help:
            command.run('help')
            print_help.assert_called_once_with()

    def test_run_with_help_and_unknown_subcommand_args_prints_help(self):
        command = core.Command()
        with patch.object(command, 'print_help') as print_help:
            command.run('help', 'invalid-subcommand-name')
            print_help.assert_called_once_with()

    def test_run_with_help_and_subcommand_args_prints_subcommand_help(self):
        command = core.Command()
        fake = command.subcommands['fake'] = Mock()
        command.run('help', 'fake')
        fake.return_value.parser.print_help.assert_called_once_with()

    def test_run_with_unknown_subcommand_arg_prints_help(self):
        command = core.Command()
        with patch.object(command, 'print_help') as print_help:
            command.run('invalid-command-name')
            print_help.assert_called_once_with()

    def test_stdout_may_be_redirected(self):
        class Fake(core.Subcommand):
            def run(self, args):
                self.stdout.write("standard output stuff")
                self.stdout.flush()
        command = core.Command()
        fake = command.subcommands['fake'] = Fake
        tempdir = tempfile.mkdtemp()
        config_path = os.path.join(tempdir, 'test.ini')
        with open(config_path, 'wt') as f:
            f.write('')
        out_path = os.path.join(tempdir, 'out.txt')
        with open(out_path, 'wt') as f:
            f.write('')
        command.run('fake', '--config', config_path, '--stdout', out_path)
        with open(out_path) as f:
            self.assertEqual(f.read(), "standard output stuff")
        shutil.rmtree(tempdir)

    def test_stderr_may_be_redirected(self):
        class Fake(core.Subcommand):
            def run(self, args):
                self.stderr.write("standard error stuff")
                self.stderr.flush()
        command = core.Command()
        fake = command.subcommands['fake'] = Fake
        tempdir = tempfile.mkdtemp()
        config_path = os.path.join(tempdir, 'test.ini')
        with open(config_path, 'wt') as f:
            f.write('')
        err_path = os.path.join(tempdir, 'err.txt')
        with open(err_path, 'wt') as f:
            f.write('')
        command.run('fake', '--config', config_path, '--stderr', err_path)
        with open(err_path) as f:
            self.assertEqual(f.read(), "standard error stuff")
        shutil.rmtree(tempdir)

    # # TODO: Figure out a better way to test this, or don't bother.
    # def test_noinit_flag_means_no_config(self):
    #     command = commands.Command()
    #     fake = command.subcommands['fake'] = Mock()
    #     command.run('fake', '--no-init')
    #     self.assertEqual(len(fake.return_value.config.files_requested), 0)


class TestSubcommand(TestCase):

    def test_add_parser_args_does_nothing(self):
        command = core.Command()
        subcommand = core.Subcommand(command)
        # Not sure this is really the way to test this, but...
        self.assertEqual(len(subcommand.parser._action_groups[0]._actions), 1)
        subcommand.add_parser_args(subcommand.parser)
        self.assertEqual(len(subcommand.parser._action_groups[0]._actions), 1)

    def test_run_not_implemented(self):
        command = core.Command()
        subcommand = core.Subcommand(command)
        args = subcommand.parser.parse_args([])
        self.assertRaises(NotImplementedError, subcommand.run, args)


# TODO: more broken tests..ugh.  these aren't very good or else i might bother
# fixing them...
# class TestFileMonitor(TestCase):

#     @patch('rattail.filemon.linux.start_daemon')
#     def test_start_daemon_with_default_args(self, start_daemon):
#         commands.main('filemon', '--no-init', 'start')
#         start_daemon.assert_called_once_with(None, None, True)

#     @patch('rattail.filemon.linux.start_daemon')
#     def test_start_daemon_with_explicit_args(self, start_daemon):
#         tmp = TempIO()
#         pid_path = tmp.putfile('test.pid', '')
#         commands.main('filemon', '--no-init', '--pidfile', pid_path, '--do-not-daemonize', 'start')
#         start_daemon.assert_called_once_with(None, pid_path, False)

#     @patch('rattail.filemon.linux.stop_daemon')
#     def test_stop_daemon_with_default_args(self, stop_daemon):
#         commands.main('filemon', '--no-init', 'stop')
#         stop_daemon.assert_called_once_with(None, None)

#     @patch('rattail.filemon.linux.stop_daemon')
#     def test_stop_daemon_with_explicit_args(self, stop_daemon):
#         tmp = TempIO()
#         pid_path = tmp.putfile('test.pid', '')
#         commands.main('filemon', '--no-init', '--pidfile', pid_path, 'stop')
#         stop_daemon.assert_called_once_with(None, pid_path)

#     @patch('rattail.commands.sys')
#     def test_unknown_platform_not_supported(self, sys):
#         tmp = TempIO()
#         stderr_path = tmp.putfile('stderr.txt', '')
#         sys.platform = 'bogus'
#         commands.main('--no-init', '--stderr', stderr_path, 'filemon', 'start')
#         sys.exit.assert_called_once_with(1)
#         with open(stderr_path) as f:
#             self.assertEqual(f.read(), "File monitor is not supported on platform: bogus\n")



# -*- coding: utf-8; -*-

from __future__ import unicode_literals, absolute_import

import os
import shutil
import tempfile

from six.moves import queue
from unittest import TestCase

from mock import Mock

from rattail.config import make_config
from rattail.filemon import linux
from rattail.filemon.config import Profile


class TestEventHandler(TestCase):

    def setUp(self):
        self.tempdir = tempfile.mkdtemp()
        self.config = make_config([])
        self.config.set(u'rattail.filemon', u'monitor', u'foo')
        self.config.set(u'rattail.filemon', u'foo.dirs', self.tempdir)
        self.config.set(u'rattail.filemon', u'foo.actions', u'noop')
        self.config.set(u'rattail.filemon', u'foo.action.noop.func', u'rattail.filemon.actions:noop')
        self.profile = Profile(self.config, u'foo')
        self.profile.queue = queue.Queue()
        self.handler = linux.EventHandler()
        self.handler.my_init(self.profile)

    def tearDown(self):
        shutil.rmtree(self.tempdir)

    def write_file(self, fname, content):
        path = os.path.join(self.tempdir, fname)
        with open(path, 'wt') as f:
            f.write(content)
        return path

    def test_in_access_event_does_nothing(self):
        event = Mock(pathname=self.write_file('file', ''))
        self.handler.process_IN_ACCESS(event)
        self.assertTrue(self.profile.queue.empty())

    def test_in_attrib_event_does_nothing(self):
        event = Mock(pathname=self.write_file('file', ''))
        self.handler.process_IN_ATTRIB(event)
        self.assertTrue(self.profile.queue.empty())

    def test_in_create_event_does_nothing(self):
        event = Mock(pathname=self.write_file('file', ''))
        self.handler.process_IN_CREATE(event)
        self.assertTrue(self.profile.queue.empty())

    def test_in_modify_event_does_nothing(self):
        event = Mock(pathname=self.write_file('file', ''))
        self.handler.process_IN_MODIFY(event)
        self.assertTrue(self.profile.queue.empty())

    def test_in_close_write_event_queues_file_if_profile_does_not_watch_locks(self):
        event = Mock(pathname=self.write_file('file', ''))
        self.profile.watch_locks = False
        self.handler.process_IN_CLOSE_WRITE(event)
        self.assertEqual(self.profile.queue.qsize(), 1)
        self.assertEqual(self.profile.queue.get_nowait(), os.path.join(self.tempdir, 'file'))

    def test_in_close_write_event_does_nothing_if_profile_watches_locks(self):
        event = Mock(pathname=self.write_file('file.lock', ''))
        self.profile.watch_locks = True
        self.handler.process_IN_CLOSE_WRITE(event)
        self.assertTrue(self.profile.queue.empty())

    def test_in_moved_to_event_queues_file_if_profile_does_not_watch_locks(self):
        event = Mock(pathname=self.write_file('file', ''))
        self.profile.watch_locks = False
        self.handler.process_IN_MOVED_TO(event)
        self.assertEqual(self.profile.queue.qsize(), 1)
        self.assertEqual(self.profile.queue.get_nowait(), os.path.join(self.tempdir, 'file'))

    def test_in_moved_to_event_does_nothing_if_profile_watches_locks(self):
        event = Mock(pathname=self.write_file('file.lock', ''))
        self.profile.watch_locks = True
        self.handler.process_IN_MOVED_TO(event)
        self.assertTrue(self.profile.queue.empty())

    def test_in_delete_event_queues_file_if_profile_watches_locks(self):
        event = Mock(pathname=self.write_file('file.lock', ''))
        self.profile.watch_locks = True
        self.handler.process_IN_DELETE(event)
        self.assertEqual(self.profile.queue.qsize(), 1)
        self.assertEqual(self.profile.queue.get_nowait(), os.path.join(self.tempdir, 'file'))

    def test_in_moved_to_event_does_nothing_if_profile_does_not_watch_locks(self):
        event = Mock(pathname=self.write_file('file', ''))
        self.profile.watch_locks = False
        self.handler.process_IN_DELETE(event)
        self.assertTrue(self.profile.queue.empty())



# -*- coding: utf-8; -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2019 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Problem Report Commands
"""

from __future__ import unicode_literals, absolute_import

from rattail.commands.core import Subcommand


class Problems(Subcommand):
    """
    Find and/or fix, and report on problems with the data
    """
    name = 'problems'
    description = __doc__.strip()

    def add_parser_args(self, parser):

        parser.add_argument('--system', '-s', dest='systems', metavar='KEY', action='append',
                            help="System for which problem checks should be performed.  You "
                            "may specify this more than once, to check multiple systems.  If "
                            "not specified, all supported systems will be checked.")

        parser.add_argument('--problem', '-p', dest='problems', metavar='KEY', action='append',
                            help="Identifies a particular problem check which should be performed.  "
                            "You may specify this more than once, to check multiple problems.  If "
                            "not specified, all supported problem reports will run.")

        parser.add_argument('--list', '-l', action='store_true',
                            help="List available problem reports, instead of running them.  "
                            "If --system args are present, will only show those reports.")

        parser.add_argument('--fix', '-F', action='store_true',
                            help="Whether to (attempt to) fix the problems, vs. just reporting them.")

        parser.add_argument('--dry-run', action='store_true',
                            help="Go through the full motions and allow logging etc. to "
                            "occur, but rollback (abort) the transaction at the end.")

    def run(self, args):
        from rattail.problems import get_problem_report_handler

        handler = get_problem_report_handler(self.config,
                                             dry_run=args.dry_run,
                                             progress=self.progress)

        # try to warn user if unknown system is specified; but otherwise ignore
        supported_systems = handler.supported_systems()
        for key in (args.systems or []):
            if key not in supported_systems:
                self.stderr.write("No problem reports exist for system: {}\n".format(key))

        reports = handler.get_problem_reports(systems=args.systems,
                                              problems=args.problems)

        if args.list:
            self.list_reports(handler, reports)
        else:
            handler.run_problem_reports(reports, fix=args.fix)

    def list_reports(self, handler, reports):
        """
        List all relevant problem reports.
        """
        organized = handler.organize_problem_reports(reports)
        for system in sorted(organized):

            self.stdout.write("\n{}\n".format(system))
            self.stdout.write("-------------------------\n")
            for problem in sorted(organized[system]):
                self.stdout.write("{}\n".format(problem))

        self.stdout.write("\n")



# -*- coding: utf-8 -*-
################################################################################
#
#  Rattail -- Retail Software Framework
#  Copyright © 2010-2017 Lance Edgar
#
#  This file is part of Rattail.
#
#  Rattail is free software: you can redistribute it and/or modify it under the
#  terms of the GNU General Public License as published by the Free Software
#  Foundation, either version 3 of the License, or (at your option) any later
#  version.
#
#  Rattail is distributed in the hope that it will be useful, but WITHOUT ANY
#  WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
#  FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
#  details.
#
#  You should have received a copy of the GNU General Public License along with
#  Rattail.  If not, see <http://www.gnu.org/licenses/>.
#
################################################################################
"""
Threading Extras
"""

from __future__ import unicode_literals

import sys
import threading


class Thread(threading.Thread):
    """
    Subclass of ``threading.Thread`` which exists only for the sake of ensuring
    the registered system exception hook (``sys.excepthook``) is invoked if an
    uncaught exception occurs within the context of the thread.
    """

    def run(self):
        try:
            super(Thread, self).run()
        except:
            sys.excepthook(*sys.exc_info())
